# build for cython functions
# Whitebeam | Kieran Molloy | Lancaster University 2020

import os

# See if Cython is installed
try:
    from Cython.Build import cythonize
# Do nothing if Cython is not available
except ImportError:
    # Got to provide this function. Otherwise, poetry will fail
    def build(setup_kwargs):
        pass
# Cython is installed. Compile
else:
    from setuptools import Extension
    from setuptools.dist import Distribution
    from distutils.command.build_ext import build_ext

    import numpy as np

    # This function will be executed in setup.py:   
    def build(setup_kwargs):

        # The file you want to compile
        extensions = Extension("whitebeam.core._whitebeam",	
                  sources=["whitebeam/core/_whitebeam.pyx"],	
                  libraries=[],	
                  include_dirs=[np.get_include()])

        # gcc arguments hack: enable optimizations
        os.environ['CFLAGS'] = '-O3'

        # Build
        setup_kwargs.update({
            'ext_modules': cythonize(
                extensions,
                language_level=3,
                compiler_directives={'linetrace': True},
            ),
            'cmdclass': {'build_ext': build_ext}
        })


# xgboosted tree
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This class implements the XGBoost base tree that appeared in:
- https://arxiv.org/abs/1603.02754 
"""

from whitebeam.core.lib import Whitebeam
import numpy as np

class XGBTree(Whitebeam):
    def __init__(self, 
                max_depth=5, 
                min_samples_split=2,
                min_samples_leaf=1,
                subsample=1.0,
                subsample_splts=1.0,
                reg_lambda=1e-2,            # regularization
                obj_tolerance=1e-2,
                random_state=1234,
                distribution="gaussian",
                n_jobs=-1,
                **kwarg):
        
        self.max_depth=max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.reg_lambda = reg_lambda
        self.distribution = distribution
        self.subsample_splts = np.clip(subsample_splts, 0.0, 1.0)
        self.obj_tolerance = obj_tolerance

        def find_split(avc):

            if avc.shape[0] == 0:
                return None

            valid_splits = np.logical_and(
                            avc[:,3] > self.min_samples_leaf,
                            avc[:,6] > self.min_samples_leaf)
            avc = avc[valid_splits,:]

            if self.subsample_splts < 1.0:
                n_avc = avc.shape[0]
                mask = np.random.rand(n_avc) < self.subsample_splts
                avc = avc[mask,:]

            if avc.shape[0] == 0:
                return None

            if self.distribution=="bernoulli":
                h_l = avc[:,5] 
                h_r = avc[:,8]
            else: 
                h_l = avc[:,3]
                h_r = avc[:,6]
            g_l = avc[:,4] 
            g_r = avc[:,7] 
            obj = g_l*g_l/(h_l+self.reg_lambda) 
            obj = obj + g_r*g_r/(h_r+self.reg_lambda)


            y_l = g_l/(h_l+self.reg_lambda)
            y_r = g_r/(h_r+self.reg_lambda)

            best_idx = np.argsort(obj)[-1]

            if (self.distribution == "bernoulli" and 
                obj[best_idx] < self.obj_tolerance):
                return None

            ss = {"selected": avc[best_idx,:],
                  "y@l": y_l[best_idx],
                  "y@r": y_r[best_idx]}

            return ss

        def is_leaf(branch, branch_parent):

            if (branch["depth"] >= self.max_depth or 
                branch["n_samples"] < self.min_samples_split):
                return True
            else:
                return False

        Whitebeam.__init__(self, 
                        find_split, 
                        is_leaf,
                        subsample=subsample, 
                        random_state=random_state,
                        n_jobs = n_jobs,
                        z_type="Hessian")

    def prune(self, X, y, y_hat, nu, max_iter=1):

        # prune status
        # - 0: don't know if it passed the test
        # - 1: passed the test

        def loss(y, y_hat):
            if self.distribution == "gaussian":
                return np.mean((y-y_hat)**2)
            elif self.distribution == "bernoulli":
                return np.mean(-2.0*(y*y_hat - np.logaddexp(0.0, y_hat)))
            else:
                return np.mean((y-y_hat)**2)

        def merge_nodes(leaf_j, leaf_k):
            leaf = leaf_j.copy()
            leaf["i_start"] = np.min([leaf_j["i_start"], 
                                        leaf_k["i_start"]])   
            leaf["i_end"] = np.max([leaf_j["i_end"], leaf_k["i_end"]]) 
            leaf["_id"] = "::".join(leaf["_id"].split("::")[:-1])
            leaf["eqs"] = leaf["eqs"][:-1]
            leaf["depth"] = leaf["depth"] - 1
            leaf["n_samples"] = leaf_j["n_samples"] + leaf_k["n_samples"]
            leaf["y_lst"] = list(leaf_j["y_lst"][:-1])
            leaf["y"] = leaf["y_lst"][-1]
            leaf["prune_status"] = 0
            return leaf

        def prune0(X, y, y_hat, nu):

            oob_mask = self.get_oob_mask()
            t = self.predict(X, "index")
            leaves = self.dump()
            sibling_pairs = self.get_sibling_pairs()

            leaves_new = []
            for j, k in sibling_pairs:

                if "prune_status" not in leaves[j]:
                    leaves[j]["prune_status"] = 0
                if k is not None and "prune_status" not in leaves[k]:
                    leaves[k]["prune_status"] = 0

                if leaves[j]["depth"] < 2:
                    leaves[j]["prune_status"] = 1
                    if k is not None and leaves[k]["depth"] < 2:
                        leaves[k]["prune_status"] = 1
                    continue

                if k is None:
                    leaves_new.append(leaves[j])
                    continue
               
                if (leaves[j]["prune_status"]==1 and
                    leaves[k]["prune_status"]==1):
                    continue

                mask_j = np.logical_and((t==j), oob_mask)
                mask_k = np.logical_and((t==k), oob_mask)
                gamma_j = leaves[j]["y"]
                gamma_k = leaves[k]["y"]

                do_merge = False
                if (np.sum(mask_j) == 0 or np.sum(mask_k) == 0):
                    do_merge = True
                else:
                    y_j = y[mask_j]
                    y_hat_j = y_hat[mask_j]
                    y_k = y[mask_k]
                    y_hat_k = y_hat[mask_k]
                    y_hat_j_new = y_hat_j + gamma_j * nu
                    y_hat_k_new = y_hat_k + gamma_j * nu

                    if loss(y_j, y_hat_j) < loss(y_j, y_hat_j_new):
                        do_merge = True
                    elif loss(y_k, y_hat_k) < loss(y_k, y_hat_k_new):
                        do_merge = True

                if do_merge:
                    leaf_m = merge_nodes(leaves[j], leaves[k])
                    mask_m = np.logical_or(mask_j, mask_k)
                    if np.sum(mask_m) > 0:
                        gamma_m = leaf_m["y"] 
                        y_m = y[mask_m]
                        y_hat_m = y_hat[mask_m]
                        y_hat_m_new = y_hat_m + gamma_m*nu
                        if loss(y_m, y_hat_m) > loss(y_m, y_hat_m_new):
                            leaf_m["prune_status"] = 1
                    leaves_new.append(leaf_m)
                else:
                    leaves[j]["prune_status"] = 1
                    leaves[k]["prune_status"] = 1
                    leaves_new.append(leaves[j])
                    leaves_new.append(leaves[k])

            for i, leaf in enumerate(leaves_new):
                leaf["index"] = i 

            self.load(leaves_new) 

            pruned = np.sum([leaf["prune_status"] for leaf in leaves_new])

            return pruned == len(leaves_new)

        if max_iter < 0:
            max_iter = np.log2(len(self.dump())) - 2

        do_pruning = True
        i = 0
        while do_pruning:
            nomore_pruning = prune0(X, y, y_hat, nu)
            i += 1
            if nomore_pruning or i >= max_iter:
                do_pruning = False

        # DONE PRUNING




# utility functions
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This module provide miscellaneous functions for the Whitebeam class such as:
- get_child_branch
- setup_canvas
- reconstruct_tree
- ...
"""

import numpy as np

PRECISION = 1e-12

def get_child_branch(ss, parent_branch, i_split, side):

    child_branch = {"eqs": list(parent_branch["eqs"])}
    ss_key = [key for key in ss.keys() if key[-2:] == side]
    ss_key += [key for key in ss.keys() if key[-2:] == "@m"]
    sidx = int(ss["selected"][0])
    svar = int(ss["selected"][1])
    sval = ss["selected"][2]
    missing = ss["selected"][9]
    parent_id = parent_branch["_id"]
    offset = 0

    if side == "@l":
        child_branch["i_start"] = parent_branch["i_start"]
        child_branch["i_end"] = i_split
        child_branch["_id"] = "{}::{}L".format(parent_id, sidx)
        child_branch["eqs"].append({"svar": svar, 
                                    "sval": sval, 
                                    #"sidx": sidx,
                                    "op": "<", 
                                    "missing": int(missing<0.5)})
    else:
        child_branch["i_start"] = i_split
        child_branch["i_end"] = parent_branch["i_end"]
        child_branch["_id"] = "{}::{}R".format(parent_id, sidx)
        child_branch["eqs"].append({"svar": svar, 
                                    "sval": sval, 
                                    #"sidx": sidx,
                                    "op": ">=", 
                                    "missing": int(missing>0.5)})
        offset = 3
   
    n_samples = ss["selected"][offset+3]
    sum_y = ss["selected"][offset+4]
    child_branch["n_samples"] = n_samples
    child_branch["y"] = sum_y / n_samples # mean(y)
    child_branch["depth"] = parent_branch["depth"] + 1

    for key in ss_key:
        key_new = key.replace(side,"")
        key_new = key_new.replace("@m","")
        child_branch[key_new] = ss[key]

    child_branch["y_lst"] = list(parent_branch["y_lst"]) 
    child_branch["y_lst"] += [child_branch["y"]]

    return child_branch

def get_xdim(X, n_hist_max):

    m = X.shape[1]

    # NOTE: x_min, x_max after removing outliers (top 1%, bottom 1%)
    x_min = np.nanpercentile(X, q=0.5, axis=0)
    x_max = np.nanpercentile(X, q=99.5, axis=0)
    #x_min = np.nanmin(np.ma.masked_invalid(X), axis=0)
    #x_max = np.nanmax(np.ma.masked_invalid(X), axis=0)

    xdim = np.zeros((m, 5), dtype=np.float, order="C")

    # Format of cavas row: [j, x_min, x_delta, n_bin, offset]
    for j in range(m):
        x_j = X[~np.isnan(X[:,j]),j]
        unique_values = np.unique(x_j)
        mask = np.logical_and(unique_values >= x_min[j], 
                            unique_values <= x_max[j])
        n_unique = np.sum(mask)
        if n_unique < n_hist_max: 
            n_bin = max(n_unique, 1)
        else:
            n_bin = n_hist_max
        xdim[j, 0] = j
        xdim[j, 1] = x_min[j]
        xdim[j, 2] = (x_max[j] - x_min[j])/n_bin # delta
        xdim[j, 3] = max(n_bin - 1, 1)
        if j > 0:
            xdim[j, 4] = xdim[j-1, 4] + xdim[j-1, 3] 

    return xdim       

def get_cnvsn(xdim): 
    # canvas for the missing values
    m, _ = xdim.shape
    cnvsn = np.zeros((m, 4))
    cnvsn[:,0] = np.arange(m)
    return cnvsn

def get_cnvs(xdim):
    m, _ = xdim.shape
    n = int(np.sum(xdim[:,3]))
    n_ = n * 2 # half for missing to the left, the other half to the right
    cnvs = np.zeros((n_, 10), dtype=np.float, order="C")
    cnvs[:,0] = np.arange(n_) 
    for j in range(m):
        x_min = xdim[j, 1]
        x_delta = xdim[j, 2]
        n_bin = int(xdim[j, 3])
        offset = int(xdim[j, 4]) * 2
        split_val = x_min
        for k in range(n_bin):
            i = offset + k
            split_val += x_delta 
            cnvs[i, 1] = j
            cnvs[i, 2] = split_val
            cnvs[(n_bin + i), 1] = j
            cnvs[(n_bin + i), 2] = split_val
            
    return cnvs

def reconstruct_tree(leaves):

    t_max = 0

    # binary search tree
    bst = {}     

    # output tree raw: [isleaf, svar, left, right, sval, out, index]
    # output tree row - 8 columns:
    #       integer section: [isleaf, svar, missing, left, right, index,
    #       float section:      sval, out]

    tree_raw = {} 
    for leaf in leaves:
        eqs = leaf["eqs"]
        node_ptr = bst 
        t = 0 # node index
        for depth, eq in enumerate(eqs):
            child_index = int(">="==eq["op"])
            svar = int(eq["svar"])
            #sidx = eq["sidx"]
            if child_index == 0:
                missing = int(eq["missing"]==0)
            else:
                missing = int(eq["missing"]==1)
            sval = float(eq["sval"])
            if "children" not in node_ptr:
                node_ptr["children"] = [{"t": t_max+1}, {"t": t_max+2}]
                tree_raw[t] = [-1, svar, missing, t_max+1, t_max+2, -1, 
                                sval, -1]
                t_max += 2
            node_ptr = node_ptr["children"][child_index]
            t = node_ptr["t"]
        tree_raw[t] = [1, -1, -1, -1, -1, leaf["index"], -1, leaf["y"]]

    tree_ind = np.zeros((t_max+1, 6), dtype=np.int, order="C")
    tree_val = np.zeros((t_max+1, 2), dtype=np.float, order="C")
    for t, node in tree_raw.items():
        tree_ind[t,:] = np.array(node[:6], dtype=np.int)
        tree_val[t,:] = np.array(node[6:], dtype=np.float)

    return tree_ind, tree_val




## just a casual empty init file :D

"""
The :mod:`whitebeam.base` module includes decision tree-based models for
classification and regression.
"""

from .alpha import AlphaTree
from .c45 import C45Tree
from .cart import DecisionTree
from .friedman import FriedmanTree
from .regr import RegTree
from .xgb import XGBTree

__all__ = ["AlphaTree",
           "C45Tree", "DecisionTree",
           "FriedmanTree", "RegTree", "XGBTree"]


# random forests
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This class implements RandomForests:
- https://www.stat.berkeley.edu/~breiman/randomforest2001.pdf
"""


from whitebeam.base import DecisionTree
import numpy as np

class RandomForests():

    def __init__(self,
                base_estimator,
                base_params,
                n_estimators=100,
                is_classifier = True):
        self.base_estimator = base_estimator
        self.base_params = base_params.copy()
        self.n_estimators = n_estimators
        self.estimators = []

        self.is_classifier = is_classifier

    def fit(self, X, y):

        X = X.astype(np.float)
        y = y.astype(np.float)
 
        whitebeam_tmp = DecisionTree()
        whitebeam_tmp.init_cnvs(X)
        xdim, cnvs, cnvsn = whitebeam_tmp.get_cnvs()
        for i in range(self.n_estimators):
            estimator = self.base_estimator(**self.base_params)
            estimator.set_cnvs(xdim, cnvs, cnvsn)
            estimator.fit(X, y, init_cnvs=False)
            self.estimators.append(estimator)

    def predict(self, X):
        n, m = X.shape
        y_hat = np.zeros(n) 
        k = len(self.estimators)
        for estimator in self.estimators:
            y_hat += estimator.predict(X) 
        y_hat /= k

        if self.is_classifier:
            return np.around(y_hat)
        else:
            return y_hat

    def dump(self, columns=[]): 
        return [estimator.dump(columns) 
                for estimator in self.estimators]
      








"""
This class implements PaloBoost, an imprvoed Stochastic Gradient TreeBoost
 that is robust to overfitting and can provide robust performance.
"""

from whitebeam.base import XGBTree

from collections import Counter

import numpy as np
from scipy.special import expit

import logging
from joblib import Parallel, delayed, cpu_count

PRECISION = 1e-5

class PaloBoost():

    def __init__(self,
                distribution="gaussian",
                learning_rate=0.1,
                subsample=0.7,
                subsample_splts=1.0,
                max_depth=3,
                n_estimators=100,
                reg_lambda=1.0,
                do_prune=True,
                n_jobs=-1,
                random_state=0,
                min_samples_split=2,
                min_samples_leaf=1):
        self.base_estimator = XGBTree
        self.n_jobs = n_jobs
        if self.n_jobs < 0:
            self.n_jobs = cpu_count()
        self.base_params = {"subsample": subsample,
                            "max_depth": max_depth,
                            "distribution": distribution,
                            "reg_lambda": reg_lambda,
                            "random_state": random_state,
                            "subsample_splts": subsample_splts,
                            "min_samples_split": min_samples_split,
                            "min_samples_leaf": min_samples_leaf,
                            "n_jobs": n_jobs}
        self.distribution = distribution
        self.nu = learning_rate
        self.n_estimators = n_estimators
        self.reg_lambda = reg_lambda
        self.do_prune = do_prune

        self.intercept = 0.0
        self.estimators = []
        self.prune_stats = []
        self.lr_stats = []
        self.feature_importances_ = None
        self.n_features_ = 0

    def fit(self, X, y):

        def initialize(y):
            if self.distribution == "gaussian":
                return np.mean(y)
            elif self.distribution == "bernoulli":
                p = np.clip(np.mean(y), PRECISION, 1-PRECISION)
                return np.log(p/(1-p))
            else:
                return np.mean(y)

        def gradient(y, y_hat):
            if self.distribution == "gaussian":
                return y - y_hat
            elif self.distribution == "bernoulli":
                return y - expit(y_hat)
            else:
                return y - y_hat

        def get_nu(y, y_hat, oob_mask, gamma, nu):

            min_n_oob = 1
            if (np.sum(oob_mask) < min_n_oob or
                np.abs(gamma) < PRECISION):
                return 0.0

            y_oob = y[oob_mask]
            y_hat_oob = y_hat[oob_mask]
           
            nu_max = nu 
            nu_min = 0 # NOTE
            if self.distribution == "gaussian": 
                nu = np.mean(y_oob - y_hat_oob)/gamma
            elif self.distribution == "bernoulli":
                num = np.sum(y_oob) + 0.5
                denom = np.sum((1-y_oob)*np.exp(y_hat_oob)) + 1.0
                nu = np.log(num/denom)/gamma
            nu = np.clip(nu, nu_min, nu_max) 

            return nu

        self.estimators = []
        self.prune_stats = []
        self.lr_stats = []

        X = X.astype(np.float)
        y = y.astype(np.float)
        if "random_state" not in self.base_params:
            self.base_params["random_state"] = 0

        n, m = X.shape 
        self.n_features_ = m
        self.feature_importances_ = np.zeros(m)
        self.intercept = initialize(y)
        
        whitebeam_tmp = self.base_estimator()
        whitebeam_tmp.init_cnvs(X)
        xdim, cnvs, cnvsn = whitebeam_tmp.get_cnvs()
        y_hat = np.full(n, self.intercept)

        with Parallel(n_jobs=self.n_jobs, prefer="threads") as parallel:
            for i in range(self.n_estimators):

                self.base_params["random_state"] += 1
                z = gradient(y, y_hat)

                estimator = self.base_estimator(**self.base_params)
                estimator.set_cnvs(xdim, cnvs, cnvsn)
                estimator.fit(X, z, init_cnvs=False, parallel=parallel)

                oob_mask = estimator.get_oob_mask()
                do_oob = estimator.is_stochastic()

                # NOTE: prune
                if do_oob and self.do_prune:
                    n_leaves_bf_prune = len(estimator.dump())
                    estimator.prune(X, y, y_hat, self.nu)
                    n_leaves_af_prune = len(estimator.dump())
                    self.prune_stats.append([i, n_leaves_bf_prune,
                                            n_leaves_af_prune])

                t = estimator.predict(X, "index")
                leaves = estimator.dump()
                avg_nu = 0.0
                for j, leaf in enumerate(leaves):
                    mask_j = (t==j)
                    gamma_j = leaf["y"]

                    # NOTE: learning-rate adjustment
                    nu_j = self.nu
                    cov_j = np.sum(mask_j)
                    if do_oob:
                        oob_mask_j = np.logical_and(mask_j, oob_mask)
                        nu_j = get_nu(y, y_hat, oob_mask_j, gamma_j, nu_j)

                    leaf["y"] = gamma_j * nu_j
                    y_hat[mask_j] += (gamma_j * nu_j)
                    avg_nu += (nu_j * cov_j)

                avg_nu = avg_nu/n
                self.lr_stats.append([i, avg_nu])
                estimator.load(leaves)
                estimator.update_feature_importances()
                self.estimators.append(estimator)

        self.update_feature_importances()

        # Done fit()

    def predict(self, X):

        n, m = X.shape
        y_hat = np.full(n, self.intercept) 

        for estimator in self.estimators:
            y_hat += estimator.predict(X) 

        if self.distribution == "bernoulli":
            y_hat = expit(y_hat)
            y_mat = np.zeros((y_hat.shape[0], 2))
            y_mat[:,0] = 1.0 - y_hat
            y_mat[:,1] = y_hat
            return y_mat
        else:
            return y_hat

    def predict_proba(self, X):
        return self.predict(X)

    def staged_predict(self, X):
        return self.staged_predict_proba(X)

    def staged_predict_proba(self, X):
        n, m = X.shape
        y_hat = np.full(n, self.intercept) 
        for stage, estimator in enumerate(self.estimators):
            y_hat += estimator.predict(X)
            if self.distribution == "bernoulli":
                y_mat = np.zeros((y_hat.shape[0],2))
                y_mat[:,1] = expit(y_hat)
                y_mat[:,0] = 1.0 - y_mat[:,1]
                yield y_mat
            else:
                yield y_hat 

    def update_feature_importances(self):
        fi = np.zeros(self.n_features_)        
        for est in self.estimators:
            fi += est.get_feature_importances()
        self.feature_importances_ = fi
        return self.feature_importances_

    def get_staged_feature_importances(self):
        fi = np.zeros(self.n_features_)        
        for i, est in enumerate(self.estimators):
            fi += est.get_feature_importances()
            yield fi

    def dump(self): 
        estimators = [estimator.dump(compact=True)
                        for estimator in self.estimators]
        return {"distribution": self.distribution,
                "learning_rate": self.nu,
                "n_estimators": self.n_estimators,
                "intercept": self.intercept, 
                "estimators": estimators}
   
    def load(self, model):
        self.distribution = model["distribution"]
        self.nu = model["learning_rate"]
        self.n_estimators = model["n_estimators"]
        self.intercept = model["intercept"] 
        for estjson in model["estimators"]:
            est = self.base_estimator()
            est.load(estjson)
            self.estimators.append(est)
        return None
 
    def get_prune_stats(self):
        return self.prune_stats

    def get_lr_stats(self):
        return self.lr_stats 





# alpha tree
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This class implements a modified version of Alpha Tree that appeared in
- the origial: https://ieeexplore.ieee.org/document/6399474/
- closer to the implementation: https://arxiv.org/abs/1606.05325
"""


from whitebeam.core.lib import Whitebeam
import numpy as np


class AlphaTree(Whitebeam):

    def __init__(self,
                alpha=1.0, 
                max_depth=5, 
                min_samples_split=2,
                min_samples_leaf=1, 
                **kwarg):

        self.alpha = alpha
        self.max_depth=max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf

        def find_split(avc):

            PRECISION = 1e-12

            if avc.shape[0] == 0:
                return None

            valid_splits = np.logical_and(
                            avc[:,3] > max(self.min_samples_leaf, 1),
                            avc[:,6] > max(self.min_samples_leaf, 1))
            avc = avc[valid_splits,:]
            if avc.shape[0] == 0:
                return None
    
            n_l = avc[:,3]
            n_r = avc[:,6]
            p_y_l = (avc[:,4] + 1.0)/(n_l + 2.0)
            p_y_r = (avc[:,7] + 1.0)/(n_r + 2.0)

            p_x_l = n_l / (n_l + n_r)
            p_x_r = n_r / (n_l + n_r)

            p_y_l[p_y_l < PRECISION] = PRECISION
            p_y_r[p_y_r < PRECISION] = PRECISION
            p_y_l[p_y_l > 1.0-PRECISION] = 1.0-PRECISION
            p_y_r[p_y_r > 1.0-PRECISION] = 1.0-PRECISION

            gain = np.zeros(avc.shape[0])

            if self.alpha == 1.0: # Information Gain
                gain += (p_x_l * p_y_l * np.log(p_y_l))
                gain += (p_x_l * (1.0-p_y_l) * np.log(1.0-p_y_l))
                gain += (p_x_r * p_y_r * np.log(p_y_r))
                gain += (p_x_r * (1.0-p_y_r) * np.log(1.0-p_y_r))
            elif self.alpha == 0.0:
                gain -= (p_x_l * np.log(p_y_l))
                gain -= (p_x_l * np.log(1.0 - p_y_l))
                gain -= (p_x_r * np.log(p_y_r))
                gain -= (p_x_r * np.log(1.0 - p_y_r))
            else:
                gain -= (p_x_l * np.power(p_y_l, self.alpha))
                gain -= (p_x_l * np.power(1.0 - p_y_l, self.alpha))
                gain -= (p_x_r * np.power(p_y_r, self.alpha))
                gain -= (p_x_r * np.power(1.0 - p_y_r, self.alpha))
                gain = gain / self.alpha / (1.0 - self.alpha)

            best_idx = np.argsort(gain)[-1]

            return {"selected": avc[best_idx,:]}

        def is_leaf(branch, branch_parent):

            if (branch["depth"] >= self.max_depth or 
                branch["n_samples"] < self.min_samples_split):
                return True
            else:
                return False

        Whitebeam.__init__(self, find_split, is_leaf, z_type="M2")




# fancy cython functions
# Whitebeam | Kieran Molloy | Lancaster University 2020

#cython: boundscheck=False
#cython: wrapround=False
#cython: cdivision=True

import numpy as np
cimport numpy as np
cimport cython
from libc.math cimport isnan

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

def reorder(X, y, z, i_start, i_end, j_split, split_value, missing):
    return _reorder(X, y, z, i_start, i_end, j_split, split_value, missing)
 
cdef size_t _reorder(
        np.ndarray[DTYPE_t, ndim=2] X, 
        np.ndarray[DTYPE_t, ndim=1] y, 
        np.ndarray[DTYPE_t, ndim=1] z, 
        size_t i_start, 
        size_t i_end, 
        size_t j_split, 
        double split_value, 
        size_t missing):
    """
    - X: 2-d numpy array (n x m)
    - y: 1-d numpy array (n)
    - z: 1-d numpy array (n)
    - i_start: row index to start
    - i_end: row index to end
    - j_split: column index for the splitting variable
    - split_value: threshold
    """
    cdef size_t j
    cdef size_t m = X.shape[1]
    cdef size_t i_head = i_start
    cdef size_t i_tail = i_end - 1
    cdef size_t do_swap = 0

    with nogil:
        while i_head <= i_tail:

            if i_tail == 0: 
                # if tail is 'zero', should break
                # otherwise, segmentation fault, 
                # as size_t has no sign. 0 - 1 => huge number
                break
            
            do_swap = 0 
            if isnan(X[i_head,j_split]):
                if missing == 1: # send the missing to the right node
                    do_swap = 1
            else:
                if X[i_head,j_split] >= split_value:
                    do_swap = 1

            if do_swap == 1:
                # swap X rows
                for j in range(m):
                    X[i_head,j], X[i_tail,j] = X[i_tail,j], X[i_head,j]
                # swap y, z values
                y[i_head], y[i_tail] = y[i_tail], y[i_head]
                z[i_head], z[i_tail] = z[i_tail], z[i_head]
                # decrease the tail index
                i_tail -= 1
            else:
                # increase the head index
                i_head += 1

    return i_head


def sketch(np.ndarray[DTYPE_t, ndim=2] X not None, 
        np.ndarray[DTYPE_t, ndim=1] y not None, 
        np.ndarray[DTYPE_t, ndim=1] z not None, 
        np.ndarray[DTYPE_t, ndim=2] xdim not None, 
        np.ndarray[DTYPE_t, ndim=2] cnvs not None, 
        np.ndarray[DTYPE_t, ndim=2] cnvsn not None):

    # canvas --> (sketch) --> avc 
    # AVC: Attribute-Value Class group in RainForest
    _sketch(X, y, z, xdim, cnvs, cnvsn)
    return 0

cdef void _sketch(
        np.ndarray[DTYPE_t, ndim=2] X, 
        np.ndarray[DTYPE_t, ndim=1] y, 
        np.ndarray[DTYPE_t, ndim=1] z, 
        np.ndarray[DTYPE_t, ndim=2] xdim, 
        np.ndarray[DTYPE_t, ndim=2] cnvs, 
        np.ndarray[DTYPE_t, ndim=2] cnvsn):

    cdef size_t i, j, k, k_raw, k_tld
    cdef size_t n = X.shape[0]
    cdef size_t m = X.shape[1]
    cdef size_t n_cnvs = <size_t> cnvs.shape[0]/2
    cdef size_t n_bin
    cdef size_t xdim0 = <size_t> xdim[0, 4]
    cdef double k_prox
    cdef double y_i, z_i
    cdef double y_tot = 0.0
    cdef double z_tot = 0.0
    cdef double n_na, y_na, z_na

    # update E[y] & E[z]
    with nogil:

        for i in range(n):

            y_i = y[i]
            z_i = z[i]
            y_tot += y_i
            z_tot += z_i

            for j in range(m):

                #if xdim[j, 2] < 1e-12:
                #    continue
                if isnan(X[i, j]):
                    cnvsn[j, 1] += 1
                    cnvsn[j, 2] += y_i
                    cnvsn[j, 3] += z_i
                else:
                    k_prox = (X[i, j] - xdim[j, 1])/xdim[j, 2]
                    if k_prox < 0:
                        k_prox = 0
                    elif k_prox > xdim[j, 3] - 1:
                        k_prox = xdim[j, 3] - 1
                    k = <size_t> (k_prox + (xdim[j, 4] - xdim0)*2)
                    cnvs[k, 3] += 1
                    cnvs[k, 4] += y_i
                    cnvs[k, 5] += z_i

        # accumulate stats
        for j in range(m):
            n_bin = <size_t> xdim[j, 3]
            
            for k_raw in range(1, n_bin): 
                k = <size_t> (k_raw + (xdim[j, 4] - xdim0)*2)
                cnvs[k, 3] += cnvs[k-1, 3] 
                cnvs[k, 4] += cnvs[k-1, 4] 
                cnvs[k, 5] += cnvs[k-1, 5] 
                # fill the right node at the same time
                cnvs[k, 6] = n - cnvs[k, 3] - cnvsn[j, 1]
                cnvs[k, 7] = y_tot - cnvs[k, 4] - cnvsn[j, 2]
                cnvs[k, 8] = z_tot - cnvs[k, 5] - cnvsn[j, 3]

            # fill the right node
            k = <size_t> ((xdim[j, 4] - xdim0)*2)
            cnvs[k, 6] = n - cnvs[k, 3] - cnvsn[j, 1]
            cnvs[k, 7] = y_tot - cnvs[k, 4] - cnvsn[j, 2]
            cnvs[k, 8] = z_tot - cnvs[k, 5] - cnvsn[j, 3]

        # missing values
        for j in range(m):

            n_bin = <size_t> xdim[j, 3]
            n_na = cnvsn[j, 1]
            y_na = cnvsn[j, 2]
            z_na = cnvsn[j, 3]

            if n_na == 0:
                continue

            for k_raw in range(n_bin):
                k = <size_t> (k_raw + (xdim[j, 4] - xdim0)*2)
                k_tld = k + n_bin
                cnvs[k_tld, 3] = cnvs[k, 3]
                cnvs[k_tld, 4] = cnvs[k, 4]
                cnvs[k_tld, 5] = cnvs[k, 5]
                cnvs[k_tld, 6] = cnvs[k, 6]
                cnvs[k_tld, 7] = cnvs[k, 7]
                cnvs[k_tld, 8] = cnvs[k, 8]
                cnvs[k_tld, 9] = 1

                cnvs[k, 3] += n_na
                cnvs[k, 4] += y_na
                cnvs[k, 5] += z_na
                cnvs[k_tld, 6] += n_na
                cnvs[k_tld, 7] += y_na
                cnvs[k_tld, 8] += z_na

    # done _sketch

def apply_tree(tree_ind, tree_val, X, y, output_type):
    if output_type == "index":
        return _apply_tree0(tree_ind, tree_val, X, y)
    else:
        return _apply_tree1(tree_ind, tree_val, X, y)

# output index
cdef np.ndarray[DTYPE_t, ndim=1] _apply_tree0(
                            np.ndarray[np.int_t, ndim=2] tree_ind, 
                            np.ndarray[DTYPE_t, ndim=2] tree_val, 
                            np.ndarray[DTYPE_t, ndim=2] X, 
                            np.ndarray[DTYPE_t, ndim=1] y):
    # Initialize node/row indicies
    cdef size_t i, t
    cdef size_t n_samples = X.shape[0]

    with nogil:
        for i in range(n_samples):
            t = 0
            while tree_ind[t,0] < 0:
                if isnan(X[i, tree_ind[t,1]]):
                    if tree_ind[t,2]==0:
                        t = tree_ind[t,3] 
                    else:
                        t = tree_ind[t,4] 
                else:
                    if X[i,tree_ind[t,1]] < tree_val[t,0]:
                        t = tree_ind[t,3]
                    else:
                        t = tree_ind[t,4]
            y[i] = tree_ind[t,5]
    return y

# output y values
cdef np.ndarray[DTYPE_t, ndim=1] _apply_tree1(
                            np.ndarray[np.int_t, ndim=2] tree_ind, 
                            np.ndarray[DTYPE_t, ndim=2] tree_val, 
                            np.ndarray[DTYPE_t, ndim=2] X, 
                            np.ndarray[DTYPE_t, ndim=1] y):
    # Initialize node/row indicies
    cdef size_t i, t
    cdef size_t n_samples = X.shape[0]

    with nogil:
        for i in range(n_samples):
            t = 0
            while tree_ind[t,0] < 0:
                if isnan(X[i, tree_ind[t,1]]):
                    if tree_ind[t,2]==0:
                        t = tree_ind[t,3] 
                    else:
                        t = tree_ind[t,4] 
                else:
                    if X[i,tree_ind[t,1]] < tree_val[t,0]:
                        t = tree_ind[t,3]
                    else:
                        t = tree_ind[t,4]
            y[i] = tree_val[t,1]
    return y





# gini tree
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This class implements the classification tree in CART 
by inheriting the alpha tree (alpha=2).
"""

from whitebeam.base.alpha import AlphaTree
import numpy as np

class DecisionTree(AlphaTree):

    def __init__(self, 
                max_depth=5, 
                min_samples_split=2,
                min_samples_leaf=1):

        AlphaTree.__init__(self, 
                        alpha=2.0,
                        max_depth=max_depth,
                        min_samples_split=min_samples_split,
                        min_samples_leaf=min_samples_leaf)





# stochastic gradient tree boost
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This class implements Stochastic Gradient TreeBoost (SGTB) in:
- https://statweb.stanford.edu/~jhf/ftp/stobst.pdf
We named it GBM (Gradient Boosting Machine), as other packages, such as
scikit-learn and R, call SGTB as GBM.
"""


from whitebeam.base import FriedmanTree
import numpy as np
from scipy.special import expit
import time
import logging
from joblib import Parallel, delayed, cpu_count

PRECISION = 1e-5

class SGTB():

    def __init__(self,
                distribution="gaussian",
                subsample=0.7,
                max_depth=3,
                learning_rate=0.1,
                n_estimators=100,
                reg_lambda=0.1,
                n_jobs=-1,
                random_state=0):
        self.base_estimator = FriedmanTree
        self.n_jobs = n_jobs
        if self.n_jobs < 0:
            self.n_jobs = cpu_count()
        self.base_params = {"subsample": subsample,
                            "max_depth": max_depth,
                            "random_state": random_state,
                            "n_jobs": self.n_jobs}
        self.distribution = distribution
        self.nu = learning_rate
        self.n_estimators = n_estimators
        self.reg_lambda = reg_lambda
        self.intercept = 0.0
        self.estimators = []
        self.feature_importances_ = None
        self.n_features_ = 0

    def fit(self, X, y):

        def initialize(y):
            if self.distribution == "gaussian":
                return np.mean(y)
            elif self.distribution == "bernoulli":
                p = np.clip(np.mean(y), PRECISION, 1-PRECISION)
                return np.log(p / (1.0 - p))
            else:
                return np.mean(y)

        def gradient(y, y_hat):
            if self.distribution == "gaussian":
                return y - y_hat
            elif self.distribution == "bernoulli":
                return y - expit(y_hat)
            else:
                return y - y_hat

        def estimate_gamma(y, y_hat):
            if self.distribution == "gaussian": 
                return np.mean(y-y_hat)
            elif self.distribution == "bernoulli":
                p = expit(y_hat)
                num = np.sum(y-p)
                denom = np.sum(p * (1-p)) + self.reg_lambda
                return num / denom
            else:
                return np.mean(y-y_hat)

        self.estimators = []

        X = X.astype(np.float)
        y = y.astype(np.float)
        if "random_state" not in self.base_params:
            self.base_params["random_state"] = 0
 
        n, m = X.shape 
        self.intercept = initialize(y)
        self.n_features_ = m
        self.feature_importances_ = np.zeros(m)

        whitebeam_tmp = self.base_estimator()
        whitebeam_tmp.init_cnvs(X)
        xdim, cnvs, cnvsn = whitebeam_tmp.get_cnvs()
        y_hat = np.zeros(n) + self.intercept
        with Parallel(n_jobs=self.n_jobs, prefer="threads") as parallel:
            for i in range(self.n_estimators):
                self.base_params["random_state"] += 1
                z = gradient(y, y_hat)
     
                estimator = self.base_estimator(**self.base_params)
                estimator.set_cnvs(xdim, cnvs, cnvsn)

                estimator.fit(X, z, init_cnvs=False, parallel=parallel)

                do_oob = estimator.is_stochastic()
                oob_mask = estimator.get_oob_mask()
                t = estimator.predict(X, "index")
                leaves = estimator.dump()

                for j, leaf in enumerate(leaves):
                    leaf_mask = (t==j)
                    mask_j = np.logical_and(leaf_mask, ~oob_mask)
                    gamma_j = estimate_gamma(y[mask_j], y_hat[mask_j])
                    leaf["_y"] = leaf["y"]
                    leaf["y"] = gamma_j * self.nu
                    y_hat[leaf_mask] += leaf["y"]

                estimator.load(leaves)
                estimator.update_feature_importances()
                self.estimators.append(estimator)

        self.update_feature_importances()

        # Done fit()

    def predict(self, X, output_type="response"):

        n, m = X.shape
        y_hat = np.full(n, self.intercept) 

        for estimator in self.estimators:
            y_hat += estimator.predict(X) 

        if self.distribution == "bernoulli":
            y_hat = expit(y_hat)
            y_mat = np.zeros((y_hat.shape[0], 2)) 
            y_mat[:,0] = 1.0 - y_hat
            y_mat[:,1] = y_hat
            return y_mat
        else:
            return y_hat

    def predict_proba(self, X):
        return self.predict(X)

    def staged_predict(self, X):
        return self.staged_predict_proba(X)

    def staged_predict_proba(self, X):

        n, m = X.shape
        y_hat = np.full(n, self.intercept) 

        for stage, estimator in enumerate(self.estimators):
            y_hat += estimator.predict(X)    
            if self.distribution == "bernoulli":
                y_mat = np.zeros((y_hat.shape[0], 2)) 
                y_mat[:,1] = expit(y_hat)
                y_mat[:,0] = 1.0 - y_mat[:,1]
                yield y_mat
            else:
                yield y_hat 

    def update_feature_importances(self):
        fi = np.zeros(self.n_features_)        
        for est in self.estimators:
            fi += est.get_feature_importances()
        self.feature_importances_ = fi
        return self.feature_importances_

    def get_staged_feature_importances(self):
        fi = np.zeros(self.n_features_)        
        for i, est in enumerate(self.estimators):
            fi += est.get_feature_importances()
            yield fi

    def dump(self, columns=[]): 
        estimators = [estimator.dump(columns)
                        for estimator in self.estimators]
        return {"intercept": self.intercept, 
                "estimators": estimators}
    
    







# main library functions
# Whitebeam | Kieran Molloy | Lancaster University 2020


"""
This module defines the Whitebeam class and its basic templates.
The Whitebeam class implments splitting data and constructing decision rules.
User need to provide two additional functions to complete the Whitebeam class:
- find_split()
- is_leaf()
"""

from whitebeam.core._whitebeam import (
    reorder, 
    sketch,
    apply_tree)
from whitebeam.utils.util import (
    reconstruct_tree,
    get_xdim,
    get_cnvsn,
    get_cnvs,
    get_child_branch)
import numpy as np
import json
from joblib import parallel_backend, Parallel, delayed, cpu_count
from scipy.special import expit
import time
import logging

class Whitebeam:
    """Base class for decision trees.
    Warning: This class should not be used directly.
    Use derived classes instead.
    lol. copied this comment from sklearn docs
    """

    def __init__(self, 
                find_split,
                is_leaf, 
                n_hist_max = 512, 
                subsample = 1.0, # subsample rate for rows (samples)
                random_state = None,
                z_type = "M2",
                n_jobs = -1,
                is_classifier=True):

        self.find_split = find_split # user-defined
        self.is_leaf = is_leaf       # user-defined
        self.n_hist_max = n_hist_max
        self.subsample = np.clip(subsample, 0.0, 1.0)
        self.random_state = random_state
        self.z_type = z_type
        self.is_classifier = is_classifier

        self.leaves = []
        self.feature_importances_ = None
        self.n_features_ = 0
        self.tree_ind   = np.zeros((1,6), dtype=np.int)
        self.tree_val   = np.zeros((1,2), dtype=np.float)
        self.mask       = None
        self.xdim       = None
        self.cnvs       = None
        self.cnvsn      = None
        self.n_jobs     = n_jobs
        if self.n_jobs < 0:
            self.n_jobs = cpu_count()

    def get_avc(self, X, y, z, i_start, i_end, parallel):
        n, m = X.shape
        y_i = y[i_start:i_end]
        z_i = z[i_start:i_end]
        self.cnvs[:,3:] = 0  # initialize canvas
        self.cnvsn[:,1:] = 0 # initialize canvas for NA

        # TODO: smart guidance on "n_jobs"
        k = int(np.ceil(m/self.n_jobs))
        def psketch(i):
            j_start = i*k
            if j_start > m-1:
                return 1
            j_end = min(m, (i+1)*k)
            jj_start = int(self.xdim[j_start,4]*2)
            jj_end = int(self.xdim[j_end-1,4]*2 + 
                        self.xdim[j_end-1,3]*2)
            X_ij = X[i_start:i_end,j_start:j_end]
            xdim_j = self.xdim[j_start:j_end,:]
            cnvs_j = self.cnvs[jj_start:jj_end,:]
            cnvsn_j = self.cnvsn[j_start:j_end,:]
            sketch(X_ij, y_i, z_i, xdim_j, cnvs_j, cnvsn_j)
            return 0

        #t0 = time.time()
        parallel(delayed(psketch)(i) for i in range(self.n_jobs))
        #t1 = time.time() - t0
        #print(i_end-i_start, t1)

        return self.cnvs

    def split_branch(self, X, y, z, branch, parallel):
        """Splits the data (X, y) into two children based on 
           the selected splitting variable and value pair.
        """

        i_start = branch["i_start"]
        i_end = branch["i_end"]
   
        # Get AVC-GROUP
        avc = self.get_avc(X, y, z, i_start, i_end, parallel)
        if avc.shape[0] < 2:
            branch["is_leaf"] = True
            return [branch]

        # Find a split SS: selected split
        ss = self.find_split(avc)     
        if (not isinstance(ss, dict) or
            "selected" not in ss):
            branch["is_leaf"] = True
            return [branch]

        svar = ss["selected"][1]
        sval = ss["selected"][2]
        missing = ss["selected"][9]

        i_split = reorder(X, y, z, i_start, i_end, 
                            svar, sval, missing)

        if i_split==i_start or i_split==i_end:
            # NOTE: this condition may rarely happen
            #       We just ignore this case, and stop the tree growth
            branch["is_leaf"] = True
            return [branch]

        left_branch = get_child_branch(ss, branch, i_split, "@l")
        left_branch["is_leaf"] = self.is_leaf(left_branch, branch)

        right_branch = get_child_branch(ss, branch, i_split, "@r")
        right_branch["is_leaf"] = self.is_leaf(right_branch, branch) 

        return [left_branch, right_branch]

    def grow_tree(self, X, y, z, branches, parallel):
        """Grows a tree by recursively partitioning the data (X, y)."""
        branches_new = []
        leaves_new = []

        for branch in branches:
            for child in self.split_branch(X, y, z, branch, parallel):
                if child["is_leaf"]:
                    leaves_new.append(child)
                else:
                    branches_new.append(child)
        return branches_new, leaves_new

    def fit(self, X, y, init_cnvs=True, parallel=None): 
        """Fit a tree to the data (X, y)."""

        n, m = X.shape
        X = X.astype(np.float, order="C", copy=True)
        y = y.astype(np.float, order="C", copy=True)
        if self.z_type=="M2":
            z = np.square(y)
        elif self.z_type=="Hessian": # bernoulli hessian
            p = expit(y) 
            z = p * (1.0 - p)
        else:
            z = np.ones(n) 
 
        if self.subsample < 1.0:
            np.random.seed(self.random_state)
            self.mask = (np.random.rand(n) < self.subsample)
            X = X[self.mask,:]
            y = y[self.mask]
            z = z[self.mask]
            n, m = X.shape
        else:
            self.mask = np.full(n, True, dtype=np.bool)

        self.n_features_ = m

        branches = [{"_id": "ROOT",
                    "is_leaf": False,
                    "depth": 0,
                    "eqs": [],
                    "i_start": 0,
                    "i_end": n,
                    "y": np.mean(y),
                    "y_lst": [], 
                    "n_samples": n}]

        if init_cnvs:
            self.init_cnvs(X)

        self.leaves = []
        if self.xdim is None or self.cnvs is None or self.cnvsn is None:
            logging.error("canvas is not initialized. no tree trained")
            return 1

        if parallel is None:
            with Parallel(n_jobs=self.n_jobs, prefer="threads") as parallel:
                while len(branches) > 0:
                    branches, leaves_new = self.grow_tree(X, y, z, 
                                                branches, parallel)
                    self.leaves += leaves_new
        else:
            # parallel-context is already given
            while len(branches) > 0:
                branches, leaves_new = self.grow_tree(X, y, z, 
                                            branches, parallel)
                self.leaves += leaves_new

        # integer index for leaves (from 0 to len(leaves))
        for i, leaf in enumerate(self.leaves): 
            leaf["index"] = i 
        self.update_feature_importances()
        self.tree_ind, self.tree_val = reconstruct_tree(self.leaves)
        return 0

    def predict(self, X, output_type="response"):
        """Predict y by applying the trained tree to X."""
        X = X.astype(np.float)
        n, m = X.shape
        y = np.zeros(n, dtype=np.float)
        out = apply_tree(self.tree_ind, self.tree_val, X, y, output_type)

        #print(out)

        # binary classifier only at the moment
        if self.is_classifier:
            return np.around(out)
        else:
            return out 

    def init_cnvs(self, X):
        self.xdim = get_xdim(X, self.n_hist_max)
        self.cnvs = get_cnvs(self.xdim)
        self.cnvsn = get_cnvsn(self.xdim)

    def set_cnvs(self, xdim, cnvs, cnvsn):
        self.xdim = xdim
        self.cnvs = cnvs
        self.cnvsn = cnvsn
        self.cnvs[:,3:] = 0  # initialize canvas
        self.cnvsn[:,1:] = 0 # initialize canvas for NA
 
    def get_cnvs(self):
        return self.xdim, self.cnvs, self.cnvsn

    def is_stochastic(self):
        return self.subsample < 1.0

    def get_mask(self):
        return self.mask
    
    def get_oob_mask(self):
        """Returns a mask array for OOB samples"""
        return ~self.mask

    def get_ttab(self):
        """Returns tree tables (ttab). 
            ttab consists of tree_ind (np_array) and tree_val (np_array).
            tree_ind stores tree indices - integer array. 
            tree_val stores node values - float array.
        """
        return self.tree_ind, self.tree_val

    def dump(self, columns=[], compact=False):
        """Dumps the trained tree in the form of array of leaves"""
        def default(o):
            if isinstance(o, np.int64): return int(o)
            raise TypeError

        n_col = len(columns)
        for leaf in self.leaves:
            for eq in leaf["eqs"]:
                if eq["svar"] < n_col:
                    eq["name"] = columns[int(eq["svar"])]
        out = json.loads(json.dumps(self.leaves, default=default))
        if compact:
            suplst = ["i_start", "i_end", "depth",
                        "_id", "n_samples", "y_lst", 
                        "is_leaf", "prune_status"] # suppress
            out_cmpct = []
            for leaf in out:
                for key in suplst:
                    leaf.pop(key, None)
                out_cmpct.append(leaf) 
            return out_cmpct
        else:
            return out

    def load(self, leaves, columns=None):
        """Loads a new tree in the form of array of leaves"""
        self.leaves = leaves
        self.tree_ind, self.tree_val = reconstruct_tree(self.leaves)
        return None

    def get_sibling_id(self, leaf_id):
        """Returns a sibling ID for the given leaf_id.
           Siblings are the nodes that are at the same level 
            with the same parent node.
        """
        sibling_id = None
        if leaf_id[-1]=="L":
            sibling_id = leaf_id[:-1] + "R"
        elif leaf_id[-1]=="R":
            sibling_id = leaf_id[:-1] + "L"
        sibling_leaf = [leaf for leaf in self.leaves 
                        if leaf["_id"]==sibling_id]
        if len(sibling_leaf) == 0:
            sibling_id = None
        return sibling_id 

    def get_sibling_pairs(self):
        """Returns an array of sibling pairs. 
            For more info, see the get_sibling_id
        """
        id2index = {leaf["_id"]:i for i, leaf in enumerate(self.leaves)}
        leaf_ids = [k for k in id2index.keys()]
        sibling_pairs = []
        while len(leaf_ids) > 0:
            leaf_id = leaf_ids.pop()
            sibling_id = self.get_sibling_id(leaf_id) 
            if sibling_id is not None:
                if sibling_id in leaf_ids:
                    leaf_ids.remove(sibling_id)
                sibling_pairs.append((id2index[leaf_id], 
                                      id2index[sibling_id]))
            else:
                sibling_pairs.append((id2index[leaf_id], None))
        return sibling_pairs
   
    def get_feature_importances(self):
        return self.feature_importances_
 
    def update_feature_importances(self):
        """Returns a modified feature importance.
            This formula takes into account of node coverage and leaf value.
            NOTE: This is different from regular feature importances that
                are used in RandomForests or GBM.
            For more info, please see the PaloBoost paper.
        """
        if self.n_features_ == 0:
            return None
        self.feature_importances_ = np.zeros(self.n_features_)
        cov = 0
        J = len(self.leaves)
        if J > 0:
            for j, leaf in enumerate(self.leaves):
                gamma_j = np.abs(leaf["y"])
                cov_j = leaf["n_samples"]
                cov += cov_j
                eff_j = cov_j*gamma_j
                for eq in leaf["eqs"]:
                    self.feature_importances_[eq["svar"]] += eff_j
            self.feature_importances_ /= J
            self.feature_importances_ /= cov
        return self.feature_importances_ 




# friedman tree
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This class implements a tree with the Friedman's splitting criterion
that appeared in: 
- https://statweb.stanford.edu/~jhf/ftp/trebst.pdf
FYI, this is the default splitting criterion for Scikit-Learn GBM.
"""


from whitebeam.core.lib import Whitebeam
import numpy as np
import logging

class FriedmanTree(Whitebeam):
    def __init__(self, 
                max_depth=5, 
                min_samples_split=2,
                min_samples_leaf=1,
                subsample=1.0,
                random_state=1234,
                n_jobs=-1,
                **kwarg):

        self.max_depth=max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
 
        def find_split(avc):

            if avc.shape[0] == 0:
                return None

            valid_splits = np.logical_and(
                            avc[:,3] > self.min_samples_leaf,
                            avc[:,6] > self.min_samples_leaf)
            avc = avc[valid_splits,:]

            if avc.shape[0] == 0:
                return None

            n_l = avc[:,3]
            n_r = avc[:,6]
            y_hat_l = avc[:,4]/n_l
            y_hat_r = avc[:,7]/n_r

            diff = y_hat_l - y_hat_r
            diff2 = diff*diff
            friedman_score = n_l * n_r / (n_l + n_r) * diff2
            
            best_idx = np.argsort(friedman_score)[-1]
            ss = {"selected": avc[best_idx,:]}

            return ss

        def is_leaf(branch, branch_parent):

            if (branch["depth"] >= self.max_depth or 
                branch["n_samples"] < self.min_samples_split):
                return True
            else:
                return False

        Whitebeam.__init__(self, 
                        find_split, 
                        is_leaf,
                        subsample=subsample, 
                        random_state=random_state,
                        n_jobs=n_jobs,
                        z_type="M2")




# regression tree
# Whitebeam | Kieran Molloy | Lancaster University 2020

"""
This class implements the regression tree in CART that uses
the minimum variance criterion.
"""

from whitebeam.core.lib import Whitebeam
import numpy as np
import logging

class RegTree(Whitebeam):
    def __init__(self, 
                max_depth=5, 
                min_samples_split=2,
                min_samples_leaf=1,
                min_varsum_decrease=0.0, 
                subsample=1.0,
                random_state=1234,
                n_jobs=-1,
                **kwarg):

        self.max_depth=max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_varsum_decrease = min_varsum_decrease

        def find_split(avc):

            if avc.shape[0] == 0:
                return None

            valid_splits = np.logical_and(
                            avc[:,3] > self.min_samples_leaf,
                            avc[:,6] > self.min_samples_leaf)
            avc = avc[valid_splits,:]

            if avc.shape[0] == 0:
                return None

            n_l = avc[:,3]
            n_r = avc[:,6]
            mu_l = avc[:,4]/n_l
            mu_r = avc[:,7]/n_r
            M2_l = avc[:,5]
            M2_r = avc[:,8]
            var_l = M2_l/n_l - mu_l*mu_l
            var_r = M2_r/n_r - mu_r*mu_r
            varsum = var_l * n_l + var_r * n_r
            best_idx = np.argsort(varsum)[0]
            best_varsum = varsum[best_idx]

            ss = {"selected": avc[best_idx,:],      # required for whitebeam
                "varsum@l": var_l[best_idx],        # required for RegTree
                "varsum@r": var_r[best_idx]}        # required for RegTree

            return ss

        def is_leaf(branch, branch_parent):

            varsum_dec = 1.0 + self.min_varsum_decrease
            if "varsum" in branch_parent:
                varsum_dec = branch_parent["varsum"] - branch["varsum"]
            if (branch["depth"] >= self.max_depth or 
                branch["n_samples"] < self.min_samples_split or
                varsum_dec < self.min_varsum_decrease):
                return True
            else:
                return False

        Whitebeam.__init__(self, 
                        find_split, 
                        is_leaf,
                        subsample = subsample, 
                        random_state = random_state,
                        n_jobs = n_jobs,
                        z_type = "M2")





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['whitebeam',
 'whitebeam.base',
 'whitebeam.core',
 'whitebeam.ensemble',
 'whitebeam.utils']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.21,<0.30.0', 'joblib>=1.0.0,<2.0.0', 'numpy>=1.19.5,<2.0.0']

setup_kwargs = {
    'name': 'whitebeam',
    'version': '1.0.3',
    'description': 'Whitebeam is a framework for creating decision tree functions.',
    'long_description': '# whitebeam\n\n| | |\n|---------|---|\n| docs    | [![Documentation Status](https://readthedocs.org/projects/whitebeam/badge/?version=latest)](https://whitebeam.readthedocs.io/en/latest/?badge=latest) |\n| test    | [![Build Status](https://travis-ci.com/K-Molloy/whitebeam.svg?branch=main)](https://travis-ci.com/K-Molloy/whitebeam) [![Requirements Status](https://requires.io/github/K-Molloy/whitebeam/requirements.svg?branch=main)](https://requires.io/github/K-Molloy/whitebeam/requirements/?branch=main)  |\n| package | [![PythonVersion](https://img.shields.io/badge/python-3.6%20%7C%203.7%20%7C%203.8-blue)](https://img.shields.io/badge/python-3.6%20%7C%203.7%20%7C%203.8-blue)  [![PyPI version](https://badge.fury.io/py/Whitebeam.svg)](https://badge.fury.io/py/Whitebeam) ![GitHub commits since latest release (by date)](https://img.shields.io/github/commits-since/K-Molloy/whitebeam/latest) ![GitHub last commit](https://img.shields.io/github/last-commit/K-Molloy/whitebeam) [![codecov](https://codecov.io/gh/K-Molloy/whitebeam/branch/main/graph/badge.svg?token=KQDZUPH74N)](https://codecov.io/gh/K-Molloy/whitebeam)|\n\n**whitebeam** is a python module for decision trees without the use of sklearn and SciPy\n\nThe project was started in 2020 by Kieran Molloy as a coursework component for SCC461 at Lancaster University\n\n# Installation\n\n## Dependencies\n\nwhitebeam requires:\n\n- Python (>= 3.6)\n- NumPy (>= 1.14)\n- Cython (>= 0.29.21)\n- joblib (>= 1.0.0)\n\n## User Installation\n\nIf you already have a working directory, the easiest installation is via `pip`\n\n```\npip install whitebeam\n```\n\nHowever it is advisable to install within a virtual environment\n\n```\nvirtualenv venv\nvenv/bin/activate\npip install whitebeam\n```\n\nThere will soon be more detailed installation instructions in the docs.\n\n# Changelog\n\nSee the changelog for a history of notable changes to whitebeam\n\n# Development\n\nDevelopment is planned to be ongoing however this may change due to university requirements.\n\n## Source Installation\n\nInstalling from source requires `python-dev` which can be installed using \n```\nsudo apt-get install python3-dev\n```\nthen cloning this repo, installing requirements and building\n```\ngit clone https://github.com/K-Molloy/whitebeam\npip install -r requirements.txt\npython setup.py build_ext --inplace\n```\nagain, it is advisable to use a virtual environment.\n\n# Testing\nAfter installation, the test suite can be launched from outside the source directory ( although `pytest `>= 5.01 must be installed)\n```\npytest whitebeam \n```\n',
    'author': 'K-Molloy',
    'author_email': 'kieran.b.molloy@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/K-Molloy/whitebeam',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
